# The MIT License (MIT)
#
# Copyright (c) 2014 Ian Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Manages pools of connections that may be reused. These pools may be shared
between threads. If is shared to a separate process, it is reset to empty.

This module should only be needed for developing new context managers, like
:class:`~provoke.mysql.MySQLConnection`, that can utilize a pool of connections
in a thread-safe manner.

"""

from __future__ import absolute_import

import os
from six.moves.queue import Queue, Empty, Full

__all__ = ['ConnectionPool']


class ConnectionPool(object):
    """Creates a connection pool for the given type of connection. The
    connection type is expected to implement ``.check()``, which returns True
    if the connection is still available and False otherwise, and ``.close()``
    which will close the connection, gracefully if possible.

    :param conn_factory: A callable object (usually a class) for creating new
                         connections when necessary.
    :type conn_factory: collections.Callable
    :param conn_args: Arguments passed to ``conn_factory``.
    :type conn_args: positional arguments
    :param conn_kwargs: Arguments passed to ``conn_factory``.
    :type conn_kwargs: keyword arguments
    :param max_size: The maximum size of the pool. Additional connections will
                     be immediately closed after use, rather than put in the
                     pool.
    :type max_size: int

    """

    #: With this static flag you can globally disable connection pooling and
    #: instead close connections immediately after use.
    enabled = True

    def __init__(self, conn_factory, conn_args=None, conn_kwargs=None,
                 max_size=10):
        super(ConnectionPool, self).__init__()
        self.conn_factory = conn_factory
        self.conn_args = conn_args or ()
        self.conn_kwargs = conn_kwargs or {}
        self.max_size = max_size
        self._pid = None
        self._get_pool()

    def __del__(self):
        pool = self._get_pool()
        while True:
            try:
                conn = pool.get_nowait()
            except Empty:
                break
            try:
                conn.close()
            except Exception:
                pass

    def _get_pool(self):
        pid = os.getpid()
        if self._pid != pid:
            self._pid = pid
            self._pool = Queue(self.max_size)
        return self._pool

    def get(self):
        """Acquires a connection. If an existing connection is available and
        ready, it is used, otherwise a new connection is established.

        :returns: A new connection object.

        """
        pool = self._get_pool()
        try:
            conn = pool.get_nowait()
            if not conn.check():
                try:
                    conn.close()
                except Exception:
                    pass
                return self.get()
            return conn
        except Empty:
            return self.conn_factory(*self.conn_args, **self.conn_kwargs)

    def release(self, conn):
        """Releases the connection back to the pool. If the pool is full, the
        connection is closed instead.

        :param conn: The connection to release.

        """
        if not self.enabled:
            conn.close()
        else:
            pool = self._get_pool()
            try:
                pool.put_nowait(conn)
            except Full:
                conn.close()




from __future__ import print_function

import sys

from provoke.app import WorkerApplication

app = WorkerApplication()
app.declare_task('do_work')

args = sys.argv[1:] or ['World!', 'Hello']
print('Sending:', ' '.join(args))
res = app.tasks.do_work.apply_async(args, send_result=True)
print('Received result:', ' '.join(res.get()))




import unittest
import os
import time
import errno
import signal
import traceback
import json
from socket import timeout as socket_timeout

from amqp.exceptions import AccessRefused

try:
    from mock import patch, MagicMock, ANY
except ImportError:
    from unittest.mock import patch, MagicMock, ANY

from provoke.amqp import AmqpConnection
from provoke.worker import _WorkerProcess, WorkerMaster, \
    get_worker_data, get_worker_app, DiscardTask


class JsonMatcher(object):

    def __init__(self, test_case, expected):
        self.test_case = test_case
        self.expected = expected

    def __eq__(self, received):
        obj = json.loads(received)
        self.test_case.assertEqual(self.expected, obj)
        return True


class TestWorkerProcess(unittest.TestCase):

    @patch.object(AmqpConnection, '__enter__')
    @patch.object(AmqpConnection, '__exit__')
    def test_consume(self, amqp_exit_mock, amqp_enter_mock):
        app = MagicMock()
        worker = _WorkerProcess(app, ['testqueue'], 1, None, None, 'exclusive')
        channel = MagicMock(callbacks=True)

        def set_done(timeout):
            channel.callbacks = False
            raise socket_timeout
        channel.connection.drain_events.side_effect = set_done
        amqp_enter_mock.return_value = channel
        worker._consume()
        amqp_enter_mock.assert_called_with()
        channel.basic_consume.assert_called_with(queue='testqueue',
                                                 consumer_tag='testqueue',
                                                 callback=ANY,
                                                 exclusive='exclusive')
        channel.connection.drain_events.assert_called_with(timeout=10.0)
        channel.connection.send_heartbeat.assert_called_with()
        amqp_exit_mock.assert_called_with(None, None, None)

    def test_send_result(self):
        channel = MagicMock()
        worker = _WorkerProcess(None, None, None, None, None, False)
        worker._send_result(channel, 'test', {})
        channel.basic_publish.assert_called_with(ANY, exchange='',
                                                 routing_key='test')

    @patch.object(_WorkerProcess, '_send_result')
    def test_on_message(self, send_result_mock):
        app = MagicMock()
        task_cb = MagicMock()
        return_cb = MagicMock()
        app.tasks.func.apply.return_value = 'return'
        worker = _WorkerProcess(app, None, None, task_cb, return_cb, False)
        worker.counter = 0
        channel = MagicMock()
        body = '{"task": "func", "args": [1], "kwargs": {"two": 2}}'
        msg = MagicMock(body=body, correlation_id='testid', reply_to='test')
        worker._on_message(channel, msg)
        task_cb.assert_called_with('func', [1], {'two': 2})
        app.tasks.func.apply.assert_called_with([1], {'two': 2}, 'testid')
        channel.basic_ack.assert_called_with(ANY)
        return_cb.assert_called_with('func', 'return')
        send_result_mock.assert_called_with(channel, 'test', ANY)
        self.assertFalse(channel.basic_cancel.called)

    def test_on_message_task_callback(self):
        app = MagicMock()
        task_cb = MagicMock(side_effect=SystemExit)
        worker = _WorkerProcess(app, None, None, task_cb, None, False)
        worker.counter = 0
        channel = MagicMock(callbacks={1: None, 2: None})
        body = '{"task": "func", "args": [1], "kwargs": {"two": 2}}'
        msg = MagicMock(body=body, correlation_id='testid')
        self.assertRaises(SystemExit, worker._on_message, channel, msg)
        task_cb.assert_called_with('func', [1], {'two': 2})
        self.assertFalse(app.tasks.func.apply.called)
        channel.basic_reject.assert_called_with(ANY, requeue=True)

    def test_on_message_task_callback_discardtask(self):
        app = MagicMock()
        task_cb = MagicMock(side_effect=DiscardTask)
        return_cb = MagicMock()
        worker = _WorkerProcess(app, None, None, task_cb, return_cb,
                                False)
        worker.counter = 0
        channel = MagicMock(callbacks={1: None, 2: None})
        body = '{"task": "func", "args": [1], "kwargs": {"two": 2}}'
        msg = MagicMock(body=body, correlation_id='testid')
        worker._on_message(channel, msg)
        task_cb.assert_called_with('func', [1], {'two': 2})
        self.assertFalse(app.tasks.func.apply.called)
        self.assertFalse(return_cb.called)
        channel.basic_ack.assert_called_with(ANY)

    def test_on_message_limit(self):
        app = MagicMock()
        worker = _WorkerProcess(app, None, 1, None, None, False)
        worker.counter = 0
        channel = MagicMock(callbacks={1: None, 2: None})

        def clear_callbacks(consumer_tag):
            channel.callbacks = {}
        channel.basic_cancel.side_effect = clear_callbacks
        body = '{"task": "func", "args": [1], "kwargs": {"two": 2}}'
        msg = MagicMock(body=body, correlation_id='testid', reply_to=None)
        worker._on_message(channel, msg)
        app.tasks.func.apply.assert_called_with([1], {'two': 2}, 'testid')
        channel.basic_ack.assert_called_with(ANY)
        channel.basic_cancel.assert_any_call(1)
        channel.basic_cancel.assert_any_call(2)

    @patch.object(_WorkerProcess, '_send_result')
    def test_on_message_task_fail(self, send_result_mock):
        app = MagicMock()
        return_cb = MagicMock()
        app.tasks.func.apply.side_effect = ValueError
        worker = _WorkerProcess(app, None, 2, None, return_cb, False)
        worker.counter = 0
        channel = MagicMock()
        body = '{"task": "func", "args": [1], "kwargs": {"two": 2}}'
        msg = MagicMock(body=body, correlation_id='testid', reply_to='test')
        self.assertRaises(ValueError, worker._on_message, channel, msg)
        app.tasks.func.apply.assert_called_with([1], {'two': 2}, 'testid')
        return_cb.assert_called_with('func', None)
        send_result_mock.assert_called_with(channel, 'test', ANY)
        channel.basic_reject.assert_called_with(ANY, requeue=False)
        self.assertFalse(channel.basic_cancel.called)

    def test_run(self):
        worker_task_callback = MagicMock()
        worker = _WorkerProcess('testapp', ['testqueue'], 10,
                                worker_task_callback, None, False)
        worker._consume = MagicMock(side_effect=SystemExit)
        worker._run()
        self.assertFalse(worker_task_callback.called)
        worker._consume.assert_called_with()

    @patch.object(time, 'sleep')
    def test_run_accessrefused(self, sleep_mock):
        worker_task_callback = MagicMock()
        worker = _WorkerProcess('testapp', ['testqueue'], 10,
                                worker_task_callback, None, False)
        worker._consume = MagicMock(side_effect=[AccessRefused, None])
        worker._run()
        self.assertFalse(worker_task_callback.called)
        sleep_mock.assert_called_once_with(ANY)
        worker._consume.assert_called_with()

    @patch.object(traceback, 'print_exc')
    def test_run_exception(self, print_exc_mock):
        worker_task_callback = MagicMock()
        worker = _WorkerProcess('testapp', ['testqueue'], 10,
                                worker_task_callback, None, False)
        worker._consume = MagicMock(side_effect=AssertionError)
        self.assertRaises(AssertionError, worker._run)
        self.assertFalse(worker_task_callback.called)
        print_exc_mock.assert_called_with()
        worker._consume.assert_called_with()


class TestWorkerMaster(unittest.TestCase):

    def test_start_callback(self):
        cb = MagicMock(side_effect=Exception)
        worker = MagicMock(queues='testqueues', pid='testpid')
        master = WorkerMaster(start_callback=cb)
        master.start_callback(worker)
        cb.assert_called_with('testqueues', 'testpid')

    def test_exit_callback(self):
        cb = MagicMock(side_effect=Exception)
        worker = MagicMock(queues='testqueues', pid='testpid')
        master = WorkerMaster('testapp', exit_callback=cb)
        master.exit_callback(worker, 'teststatus')
        cb.assert_called_with('testqueues', 'testpid', 'teststatus')

    def test_add_worker(self):
        master = WorkerMaster()
        self.assertEqual([], master.workers)
        master.add_worker('testapp', ['queue1'], exclusive=True)
        self.assertEqual(1, len(master.workers))
        self.assertEqual('testapp', master.workers[0].app)
        self.assertEqual(['queue1'], master.workers[0].queues)
        self.assertEqual(None, master.workers[0].pid)
        master.add_worker('testapp', ['queue2'], num_processes=2)
        self.assertEqual(3, len(master.workers))
        self.assertEqual('testapp', master.workers[0].app)
        self.assertEqual(['queue1'], master.workers[0].queues)
        self.assertTrue(master.workers[0].exclusive)
        self.assertEqual('testapp', master.workers[1].app)
        self.assertEqual(['queue2'], master.workers[1].queues)
        self.assertFalse(master.workers[1].exclusive)
        self.assertEqual('testapp', master.workers[2].app)
        self.assertEqual(['queue2'], master.workers[2].queues)
        self.assertFalse(master.workers[2].exclusive)

    @patch.object(os, 'waitpid')
    def test_check_workers(self, waitpid_mock):
        exit_cb = MagicMock()
        worker = MagicMock(queues='testqueues', pid='testpid')
        master = WorkerMaster('testapp', exit_callback=exit_cb)
        master.workers = [worker]
        waitpid_mock.return_value = ('testpid', 0)
        self.assertTrue(master._check_workers())
        waitpid_mock.assert_called_with(0, 0)
        exit_cb.assert_called_with('testqueues', 'testpid', 0)
        self.assertEqual(None, worker.pid)

    @patch.object(os, 'waitpid')
    def test_check_workers_no_children(self, waitpid_mock):
        exit_cb = MagicMock()
        worker = MagicMock(queues='testqueues', pid='testpid')
        master = WorkerMaster('testapp', exit_callback=exit_cb)
        master.workers = [worker]
        waitpid_mock.side_effect = OSError(errno.ECHILD, 'No child processes')
        self.assertFalse(master._check_workers())
        waitpid_mock.assert_called_with(0, 0)
        exit_cb.assert_called_with('testqueues', 'testpid', None)
        self.assertEqual(None, worker.pid)

    @patch.object(os, 'waitpid')
    def test_check_workers_other_oserror(self, waitpid_mock):
        exit_cb = MagicMock()
        worker = MagicMock(queues='testqueues', pid='testpid')
        master = WorkerMaster('testapp', exit_callback=exit_cb)
        master.workers = [worker]
        waitpid_mock.side_effect = OSError(99, 'Something else')
        self.assertRaises(OSError, master._check_workers)
        waitpid_mock.assert_called_with(0, 0)
        self.assertEqual('testpid', worker.pid)

    @patch.object(os, 'fork')
    @patch.object(os, '_exit')
    def test_start_worker_child(self, exit_mock, fork_mock):
        app = MagicMock()
        worker = _WorkerProcess(app, ['testqueue'], None, None, None, False)
        worker._run = MagicMock()
        master = WorkerMaster('testapp')
        fork_mock.return_value = 0
        master._start_worker(worker)
        fork_mock.assert_called_with()
        worker._run.assert_called_with()
        exit_mock.assert_called_with(os.EX_OK)

    @patch.object(os, 'fork')
    def test_start_worker_parent(self, fork_mock):
        worker = MagicMock()
        master = WorkerMaster('testapp')
        fork_mock.return_value = 13
        self.assertEqual(13, master._start_worker(worker))
        fork_mock.assert_called_with()

    @patch.object(os, 'fork')
    @patch.object(os, '_exit')
    def test_get_worker_data(self, exit_mock, fork_mock):
        app = MagicMock()
        worker = _WorkerProcess(app, ['testqueue'], None, None, None, False)
        worker._run = MagicMock()
        master = WorkerMaster('testapp', worker_data={'test': 'data'})
        fork_mock.return_value = 0
        self.assertRaises(RuntimeError, get_worker_data, 'test')
        master._start_worker(worker)
        self.assertEquals('data', get_worker_data('test'))
        fork_mock.assert_called_with()
        worker._run.assert_called_with()
        exit_mock.assert_called_with(os.EX_OK)

    @patch.object(os, 'fork')
    @patch.object(os, '_exit')
    def test_get_worker_app(self, exit_mock, fork_mock):
        app = MagicMock()
        worker = _WorkerProcess(app, ['testqueue'], None, None, None, False)
        worker._run = MagicMock()
        master = WorkerMaster('testapp')
        fork_mock.return_value = 0
        self.assertRaises(RuntimeError, get_worker_app)
        master._start_worker(worker)
        self.assertEquals(app, get_worker_app())
        fork_mock.assert_called_with()
        worker._run.assert_called_with()
        exit_mock.assert_called_with(os.EX_OK)

    def test_restart_workers(self):
        start_cb = MagicMock()
        worker1 = MagicMock(queues='testqueues', pid='testpid1')
        worker2 = MagicMock(queues='testqueues', pid=None)
        master = WorkerMaster(start_callback=start_cb)
        master._start_worker = MagicMock(return_value='testpid2')
        master.workers = [worker1, worker2]
        master._restart_workers()
        master._start_worker.assert_called_once_with(worker2)
        start_cb.assert_called_with('testqueues', 'testpid2')

    @patch.object(os, 'kill')
    def test_stop_workers(self, kill_mock):
        kill_mock.side_effect = [None,
                                 OSError(errno.ESRCH, 'No such pid'),
                                 OSError(99, 'Something else')]
        worker1 = MagicMock(pid='testpid1')
        worker2 = MagicMock(pid='testpid2')
        worker3 = MagicMock(pid='testpid3')
        master = WorkerMaster('testapp')
        master.workers = [worker1, worker2, worker3]
        self.assertRaises(OSError, master._stop_workers)
        kill_mock.assert_any_call('testpid1', signal.SIGTERM)
        kill_mock.assert_any_call('testpid2', signal.SIGTERM)
        kill_mock.assert_any_call('testpid3', signal.SIGTERM)

    @patch.object(os, 'waitpid')
    def test_wait(self, waitpid_mock):
        exit_cb = MagicMock()
        waitpid_mock.side_effect = [('testpid1', 0),
                                    OSError(errno.ESRCH, 'No such pid'),
                                    OSError(99, 'Something else')]
        worker1 = MagicMock(queues=None, pid='testpid1')
        worker2 = MagicMock(queues=None, pid='testpid2')
        worker3 = MagicMock(queues=None, pid='testpid3')
        master = WorkerMaster('testapp', exit_callback=exit_cb)
        master.workers = [worker1, worker2, worker3]
        self.assertRaises(OSError, master.wait)
        waitpid_mock.assert_any_call('testpid1', 0)
        waitpid_mock.assert_any_call('testpid2', 0)
        waitpid_mock.assert_any_call('testpid3', 0)
        exit_cb.assert_any_call(None, 'testpid1', 0)
        exit_cb.assert_any_call(None, 'testpid2', None)

    def test_run(self):
        master = WorkerMaster('testapp')
        master._restart_workers = MagicMock(side_effect=[None, ValueError])
        master._check_workers = MagicMock()
        master._stop_workers = MagicMock()
        self.assertRaises(ValueError, master.run)
        master._restart_workers.assert_any_call()
        master._restart_workers.assert_called_with()
        master._check_workers.assert_called_with()
        master._stop_workers.assert_called_with()




import unittest

try:
    from mock import patch, MagicMock
except ImportError:
    from unittest.mock import patch, MagicMock

from provoke.connectionpool import ConnectionPool


class TestConnectionPool(unittest.TestCase):

    def test_get_new(self):
        conn_factory = MagicMock()
        conn_factory.return_value = conn = MagicMock()
        pool = ConnectionPool(conn_factory, (1, 2), {'three': 4})
        self.assertEqual(conn, pool.get())
        conn_factory.assert_called_with(1, 2, three=4)

    def test_get_existing(self):
        conn = MagicMock()
        conn.check.return_value = True
        pool = ConnectionPool(None)
        pool._get_pool().put_nowait(conn)
        self.assertEqual(conn, pool.get())
        conn.check.assert_called_with()

    def test_get_existing_broken(self):
        conn_factory = MagicMock()
        conn_factory.return_value = new_conn = MagicMock()
        old_conn = MagicMock()
        old_conn.check.return_value = False
        pool = ConnectionPool(conn_factory)
        pool._get_pool().put_nowait(old_conn)
        self.assertEqual(new_conn, pool.get())
        conn_factory.assert_called_with()
        old_conn.check.assert_called_with()

    def test_release(self):
        conn = MagicMock()
        pool = ConnectionPool(None)
        pool.release(conn)
        self.assertEqual(1, pool._get_pool().qsize())
        self.assertFalse(conn.close.called)

    def test_release_full(self):
        conn1 = MagicMock()
        conn2 = MagicMock()
        pool = ConnectionPool(None, max_size=1)
        pool.release(conn1)
        pool.release(conn2)
        self.assertEqual(1, pool._get_pool().qsize())
        self.assertFalse(conn1.close.called)
        conn2.close.assert_called_with()

    @patch.object(ConnectionPool, 'enabled', new=False)
    def test_release_disabled(self):
        conn = MagicMock()
        pool = ConnectionPool(None)
        pool.release(conn)
        conn.close.assert_called_with()



# The MIT License (MIT)
#
# Copyright (c) 2014 Ian Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Loads plugins defining worker processes and task functions. Starts and
manages the pool of worker processes to execute tasks as the come in.

"""

from __future__ import absolute_import, print_function

import sys
import time
import logging
import signal
import resource
from optparse import OptionParser

from pkg_resources import iter_entry_points
from six.moves.configparser import SafeConfigParser

from ..config import Configuration
from . import system, WorkerMaster


logger = logging.getLogger('provoke.main')


class BadPlugin(Exception):
    pass


class ReloadSignal(Exception):
    pass


def handle_signals():
    def exit_sig(signum, frame):
        sys.exit(0)

    def reload_sig(signum, frame):
        raise ReloadSignal()
    signal.signal(signal.SIGTERM, exit_sig)
    signal.signal(signal.SIGHUP, reload_sig)


def start_master(options, plugins):
    cfgparser = SafeConfigParser()
    cfgparser.read(options.config)
    cfg = Configuration(cfgparser)

    pidfile = None
    user, group, umask = None, None, None

    def process_init():
        system.drop_privileges(user, group, umask)
        time.sleep(0.1)

    master = WorkerMaster(process_callback=process_init)

    cfg.configure_logging()
    cfg.configure_taskgroups()
    cfg.configure_amqp()
    cfg.configure_mysql()

    for plugin in plugins:
        for entry_point in iter_entry_points('provoke.workers', plugin):
            logger.debug('Loading plugin: name=%s', entry_point.name)
            register = entry_point.load()
            register(master, cfgparser)
            break
        else:
            raise BadPlugin(plugin)

    for res, limits in cfg.get_rlimits():
        resource.setrlimit(res, limits)

    if options.daemon:
        logger.debug('Daemonizing master process')
        pidfile = cfg.get_pidfile()
        stdout, stderr, stdin = cfg.get_stdio_redirects()
        user, group, umask = cfg.get_worker_privileges()
        if stdout or stderr or stdin:
            system.redirect_stdio(stdout, stderr, stdin)
        system.daemonize()
    with system.PidFile(pidfile):
        handle_signals()
        try:
            master.run()
        finally:
            master.wait()


def main():
    usage = 'Usage: %prog [options] <plugins>'
    parser = OptionParser(description=__doc__, usage=usage)
    parser.add_option('--config', action='append', metavar='PATH', default=[],
                      help='Configuration file')
    parser.add_option('--daemon', action='store_true', default=False,
                      help='Daemonize the master process')
    parser.add_option('-l', '--list-plugins', action='store_true',
                      default=False, help='List all known plugins')

    options, plugins = parser.parse_args()

    if options.list_plugins:
        for entry_point in iter_entry_points('provoke.workers'):
            print(entry_point.name)
        return
    elif not plugins:
        parser.error('One or more plugins required')

    while True:
        try:
            start_master(options, plugins)
        except BadPlugin as exc:
            parser.error('Unrecognized plugin name: ' + str(exc))
        except ReloadSignal:
            pass
        except (SystemExit, KeyboardInterrupt):
            break



# The MIT License (MIT)
#
# Copyright (c) 2014 Ian Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Common HTTP module. Provides a central point where HTTP endpoints may be
configured, and a context manager to ensure connections are closed after use.

"""

from __future__ import absolute_import

from six.moves import http_client
import base64

__all__ = ['HttpConnection']


class HttpConnection(object):
    """Defines the context manager for connecting to an HTTP endpoint and
    cleaning up afterwards. The resulting connection is a :py:mod:`http_client`
    object. For example::

        with HttpConnection('example-endpoint') as conn, headers:
            conn.request('GET', '/', headers=headers)
            res = conn.getresponse()
            assert res.status == 200

    :param name: The pre-configured endpoint name.
    :type name: str

    """

    _connection_params = {}

    def __init__(self, name):
        super(HttpConnection, self).__init__()
        self.name = name
        self.conn = None

    @classmethod
    def reset_connection_params(cls):
        """Removes all existing connection parameter information."""
        cls._connection_params = {}

    @classmethod
    def set_connection_params(cls, name, **params):
        """Configures connection parameters that future usages of the context
        manager will use to connect to the HTTP endpoint.

        :param name: A unique, memorable name to reference the endpoint by.
        :type name: str
        :param params: The HTTP endpoint parameters. The boolean keyword
                       ``ssl`` determines whether the rest of the keywords will
                       be passed into :py:class:`http_client.HTTPConnection` or
                       :py:class:`http_client.HTTPSConnection`. The keywords
                       ``user`` and ``password`` will be used to construct a
                       Basic *Authorization* header.
        :type params: Keyword arguments

        """
        cls._connection_params[name] = params

    def _create_conn(self, params):
        user = params.pop('user', None)
        password = params.pop('password', None)
        host = params.pop('host', 'localhost')
        if params.pop('ssl', False):
            conn = http_client.HTTPSConnection(host, **params)
        else:
            conn = http_client.HTTPConnection(host, **params)
        headers = {}
        if user is not None:
            auth = '{0}:{1}'.format(user, password).encode('utf-8')
            auth_raw = base64.b64encode(auth)
            headers['Authorization'] = b'Basic ' + auth_raw
        return conn, headers

    def __enter__(self):
        params = self._connection_params.get(self.name, {}).copy()
        self.conn, headers = self._create_conn(params)
        return self.conn, headers

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.conn.close()
        except Exception:
            if not exc_type:
                raise




"""Contains functions to simplify the usual daemonization procedures for long-
running processes.

"""

from __future__ import absolute_import

import os
import os.path
import sys
from pwd import getpwnam
from grp import getgrnam

__all__ = ['daemonize', 'redirect_stdio', 'drop_privileges', 'PidFile']


def daemonize():
    """Daemonizes the current process using the standard double-fork.
    This function does not affect standard input, output, or error.

    :returns: The PID of the daemonized process.

    """

    # Fork once.
    try:
        pid = os.fork()
        if pid > 0:
            os._exit(0)
    except OSError:
        return

    # Set some options to detach from the terminal.
    os.chdir('/')
    os.setsid()
    os.umask(0)

    # Fork again.
    try:
        pid = os.fork()
        if pid > 0:
            os._exit(0)
    except OSError:
        return

    os.setsid()
    return os.getpid()


def redirect_stdio(stdout=None, stderr=None, stdin=None):
    """Redirects standard output, error, and input to the given
    filenames. Standard output and error are opened in append-mode, and
    standard input is opened in read-only mode. Leaving any parameter
    blank leaves that stream alone.

    :param stdout: filename to append the standard output stream into.
    :param stderr: filename to append the standard error stream into.
    :param stdin: filename to read from as the standard input stream.

    """

    # Find the OS /dev/null equivalent.
    nullfile = getattr(os, 'devnull', '/dev/null')

    # Redirect all standard I/O to /dev/null.
    sys.stdout.flush()
    sys.stderr.flush()
    si = open(stdin or nullfile, 'r')
    so = open(stdout or nullfile, 'a+')
    se = open(stderr or nullfile, 'a+', 0)
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    # Set permissions as necessary.
    if stdin:
        os.fchmod(si.fileno(), 0o600)
    if stdout:
        os.fchmod(so.fileno(), 0o600)
    if stderr:
        os.fchmod(se.fileno(), 0o600)


def drop_privileges(user=None, group=None, mask=None):
    """Uses system calls to drop privileges to the given user and group.  This
    is useful for security purposes, once root-only ports like 25 are opened.

    .. seealso:: :func:`os.setuid`, :func:`os.setgid`, :func:`os.umask`

    :param user: user name (from /etc/passwd) or UID.
    :param group: group name (from /etc/group) or GID.
    :param mask: file mode creation mask.

    """
    if group:
        try:
            gid = int(group)
        except ValueError:
            gid = getgrnam(group).gr_gid
        os.setgid(gid)
    if user:
        try:
            uid = int(user)
        except ValueError:
            uid = getpwnam(user).pw_uid
        os.setuid(uid)
    if mask:
        os.umask(mask)


class PidFile(object):
    """Context manager which creates a PID file containing the current process
    id, runs the context, and then removes the PID file.

    An :py:exc:`OSError` exceptions when creating the PID file will be
    propogated without executing the context.

    :param filename: The filename to use for the PID file. If ``None`` is
                     given, the context is simply executed with no PID file
                     created.

    """

    def __init__(self, filename=None):
        super(PidFile, self).__init__()
        if not filename:
            self.filename = None
        else:
            self.filename = os.path.abspath(filename)

    def __enter__(self):
        if self.filename:
            with open(self.filename, 'w') as pid:
                pid.write('{0}\n'.format(os.getpid()))
            return self.filename

    def __exit__(self, exc_type, exc_value, traceback):
        if self.filename:
            try:
                os.unlink(self.filename)
            except OSError:
                pass



# The MIT License (MIT)
#
# Copyright (c) 2014 Ian Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Common worker application module. Provides the interface with which
producers and executors see the task system.

.. class:: _TaskCaller

   This object represents a task that may be called synchronously in the
   current process, or asynchronously by a separate worker process.

   .. method:: delay(*args, **kwargs)

      Convenience method for calling :meth:`.apply_async` in a more
      traditional fashion.

      :param args: Positional arguments passed in to the task function.
      :type args: positional arguments
      :param kwargs: Keyword arguments passed in to the task function.
      :type kwargs: keyword arguments
      :returns: A way to retrieve the result of the task's execution when it
                is ready.
      :rtype: :class:`AsyncResult`

   .. method:: apply_async(args, kwargs=None, correlation_id=None,
                           routing_key=None, send_result=False)

      Triggers the asynchronous execution of a task by a separate worker
      process. Tasks will be executed in the order they are triggered, so a
      task may not be executed immediately.

      :param args: Positional arguments to be passed in to the task function.
      :type args: list
      :param kwargs: Keyword arguments to be passed in to the task function.
      :type kwargs: dict
      :param correlation_id: Force the message to use a specific correlation
                             identifier string.
      :type correlation_id: str
      :param routing_key: Override the default AMQP routing key for the task
                          with the given string.
      :type routing_key: str
      :param send_result: Create a temporary result queue and request that the
                          worker publish the task's result to it.
      :type send_result: bool
      :returns: If ``send_result`` is True, used retrieve the result of the
                task's execution when it is ready. Otherwise, returns ``None``.
      :rtype: :class:`AsyncResult`

   .. method:: apply(args, kwargs=None, correlation_id=None)

      Calls the task function synchronously in the current process, returning
      the result. This is equivalent to a normal function call, except
      start and finish events are logged.

      :param args: Arguments passed directly into the function.
      :type args: list
      :param kwargs: Arguments passed directly into the function.
      :type kwargs: dict
      :param correlation_id: The task being executed had this string as its
                             correlation identifier.
      :type correlation_id: str
      :returns: The return value of the function call.

   .. method:: __call__(*args, **kwargs)

      Equivalent to :meth:`.apply` passing in ``args`` and ``kwargs``.

      :param args: Arguments passed directly into the function.
      :type args: positional arguments
      :param kwargs: Arguments passed directly into the function.
      :type kwargs: keyword arguments
      :returns: The return value of the function call.

"""

from __future__ import absolute_import

import time
import json
import errno
import logging
from uuid import uuid4
from socket import timeout as socket_timeout, error as socket_error
from multiprocessing import TimeoutError

from six.moves import cPickle
import amqp

from .amqp import AmqpConnection

__all__ = ['WorkerApplication', 'taskgroup']


logger = logging.getLogger('provoke.tasks')


class AsyncResult(object):
    """Used to query for the results of an asynchronous operation. Compatible
    with the builtin :py:class:`multiprocessing.pool.AsyncResult` class
    interface.

    :param correlation_id: The ID used to identify tasks and correlate their
                           request and response messages.
    :type correlation_id: str

    """

    def __init__(self, correlation_id):
        super(AsyncResult, self).__init__()
        self.correlation_id = correlation_id

    @property
    def result_queue(self):
        """The name of the queue where results will be published."""
        return 'result_{0}'.format(self.correlation_id)

    @property
    def args(self):
        """If the result is available, these will be a tuple of the original
        positional and keyword arguments sent with the request.

        """
        if self.ready():
            return (self._result['args'], self._result['kwargs'])
        raise AttributeError

    @property
    def name(self):
        """If the result is available, this will be the string name of the
        completed task.

        """
        if self.ready():
            return self._result['task_name']
        raise AttributeError

    @property
    def returned(self):
        """If the result is available and was successful, this will be the
        returned value of the completed task.

        """
        if self.ready():
            if hasattr(self, '_return'):
                return self._return
        raise AttributeError

    @property
    def exception(self):
        """If the result is available and was not successful, this will be the
        exception object raised by the completed task.

        """
        if self.ready():
            if hasattr(self, '_exc'):
                return self._exc
        raise AttributeError

    @property
    def traceback(self):
        """If the result is available and was not successful, this will be the
        exception traceback from the completed task.

        """
        if self.ready():
            if hasattr(self, '_exc'):
                return self._result['exception']['traceback']
        raise AttributeError

    def _get_cached_result(self):
        if hasattr(self, '_return'):
            return self._return
        elif hasattr(self, '_exc'):
            raise self._exc

    def _on_message(self, msg):
        self._result = res = json.loads(msg.body)
        if 'return' in res:
            self._return = res['return']
        elif 'exception' in res:
            exception_raw = res['exception']['value'].encode('ascii')
            self._exc = cPickle.loads(exception_raw)

    def _check(self):
        with AmqpConnection() as channel:
            try:
                msg = channel.basic_get(queue=self.result_queue, no_ack=True)
            except amqp.exceptions.NotFound:
                raise KeyError(self.result_queue)
            if msg:
                self._on_message(msg)
                channel.queue_delete(self.result_queue)
                return self._get_cached_result()
        raise TimeoutError(0.0)

    def _wait(self, timeout):
        start_time = time.time()
        with AmqpConnection() as channel:
            try:
                channel.basic_consume(queue=self.result_queue,
                                      no_ack=True, callback=self._on_message)
            except amqp.exceptions.NotFound:
                raise KeyError(self.result_queue)
            while channel.callbacks:
                elapsed = time.time() - start_time
                cur_timeout = 10.0
                if timeout is not None:
                    remaining = timeout - elapsed
                    cur_timeout = max(min(remaining, 10.0), 0.0)
                try:
                    channel.connection.drain_events(timeout=cur_timeout)
                except socket_timeout:
                    channel.connection.send_heartbeat()
                except socket_error as exc:
                    if exc.errno != errno.EAGAIN:
                        raise
                if hasattr(self, '_result'):
                    channel.queue_delete(self.result_queue)
                    return self._get_cached_result()
                if timeout is not None and remaining <= 0.0:
                    break
        raise TimeoutError(timeout)

    def get(self, timeout=None):
        """Returns the task result when it becomes available. If the result was
        an exception, the exception is re-raised by this method.

        :param timeout: If this many seconds elapse before the result is ready,
                        this method will stop and return ``None``. The default
                        is to wait indefinitely.
        :type timeout: float
        :raises: :py:exc:`~multiprocessing.TimeoutError`

        """
        if hasattr(self, '_result'):
            return self._get_cached_result()
        if timeout == 0.0:
            return self._check()
        else:
            return self._wait(timeout)

    def wait(self, timeout=None):
        """Waits for the task result to become available. It does not return
        the result or re-raise any exceptions.

        :param timeout: Wait at most this many seconds.
        :type timeout: float

        """
        if hasattr(self, '_result'):
            return
        try:
            self.get(timeout)
        except Exception:
            pass

    def ready(self):
        """Checks if the result is immediately available. If so, the
        :meth:`.get` and :meth:`.wait` methods will return immediately.

        :rtype: bool

        """
        if hasattr(self, '_result'):
            return True
        self.wait(0.0)
        return hasattr(self, '_result')

    def successful(self):
        """Checks if the result is both available and the task result will not
        re-raise an exception.

        :rtype: bool

        """
        if self.ready():
            return hasattr(self, '_return')
        return False


class _TaskCaller(object):

    def __init__(self, func, name, app, exchange, routing_key):
        super(_TaskCaller, self).__init__()
        self.func = func
        self.name = name
        self.app = app
        self.exchange = exchange or ''
        if routing_key is None:
            self.routing_key = name
        else:
            self.routing_key = routing_key

    def delay(self, *args, **kwargs):
        return self.apply_async(args, kwargs)

    def apply_async(self, args, kwargs=None, correlation_id=None,
                    routing_key=None, send_result=False):
        if correlation_id is None:
            correlation_id = str(uuid4())
        job = {'task': self.name,
               'args': args,
               'kwargs': kwargs}
        job_raw = json.dumps(job)
        result = AsyncResult(correlation_id)
        reply_to = result.result_queue if send_result else None
        msg = amqp.Message(job_raw,
                           content_type='application/json',
                           reply_to=reply_to,
                           correlation_id=correlation_id)
        if routing_key is None:
            routing_key = self.routing_key
        with AmqpConnection() as channel:
            if send_result:
                channel.queue_declare(queue=reply_to,
                                      auto_delete=False)
            channel.basic_publish(msg, exchange=self.exchange,
                                  routing_key=routing_key)
        logger.info('Task queued: name=%s, id=%s', self.name, correlation_id)
        logger.debug('Task details: args=%s, kwargs=%s',
                     repr(args), repr(kwargs))
        if send_result:
            return result

    def apply(self, args, kwargs=None, correlation_id=None):
        logger.info('Task starting: name=%s, id=%s', self.name, correlation_id)
        kwargs = kwargs or {}
        start_time = time.time()
        try:
            return self(*args, **kwargs)
        finally:
            elapsed = time.time() - start_time
            logger.info('Task finished: name=%s, id=%s, elapsed=%s',
                        self.name, correlation_id, elapsed)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)


class _TaskSet(object):

    def __init__(self, app):
        super(_TaskSet, self).__init__()
        self._app = app
        self._tasks = {}

    def _set(self, func, name, exchange, routing_key):
        call = _TaskCaller(func, name, self._app, exchange, routing_key)
        self._tasks[name] = call

    def _declare(self, name, exchange, routing_key):
        call = _TaskCaller(None, name, self._app, exchange, routing_key)
        self._tasks.setdefault(name, call)

    def __len__(self):
        return len(self._tasks)

    def __contains__(self, name):
        return name in self._tasks

    def __getattr__(self, name):
        call = self._tasks.get(name)
        if call:
            return call
        raise AttributeError(name)

    def __repr__(self):
        task_names = self._tasks.keys()
        return '<registered task set {0!r}>'.format(task_names)


def taskgroup(name):
    """Convenience decorator for ensuring that the task is a member of the
    given taskgroup.

    :param name: The taskgroup to make the task a member of.
    :type name: str

    """
    def deco(func):
        func._taskgroup = name
        return func
    return deco


class WorkerApplication(object):
    """Defines an application that has a set of tasks that may be executed
    asynchronously by worker processes.

    """

    _taskgroups = {None: {'exchange': '', 'routing_key': None}}

    def __init__(self):
        super(WorkerApplication, self).__init__()

        #: This attribute should be used by clients to call tasks by name. For
        #: example, to call a task registered as ``'do_stuff'``::
        #:
        #:     tasks.do_stuff.delay(3.14159, 'arg2')
        #:
        #: Tasks in this object are objects of type :class:`_TaskCaller` and
        #: so they may be executed asynchronously with
        #: :meth:`~_TaskCaller.delay` and :meth:`~_TaskCaller.apply_async` or
        #: synchronously by calling them like a normal function.
        self.tasks = _TaskSet(self)

    @classmethod
    def reset_taskgroups(cls):
        """Removes all known taskgroups."""
        cls._taskgroups = {None: {'exchange': '', 'routing_key': None}}

    @classmethod
    def declare_taskgroup(cls, name, exchange='', routing_key=None):
        """Associates a name with a set of routing information. Tasks that are
        a member of a given taskgroup will use its routing information. Tasks
        that are not a member of a taskgroup will raise a ``TypeError``
        exception if they are called asynchronously. Taskgroups are shared
        between all instances of this class.

        :param name: The name of the taskgroup.
        :type name: str
        :param exchange: The AMQP exchange name to use for routing the task
                         message.
        :type exchange: str
        :param routing_key: The AMQP routing key used for routing the task
                            message.  If this value is ``None``, the task name
                            will be used instead.
        :type routing_key: str

        """
        cls._taskgroups[name] = {'exchange': exchange,
                                 'routing_key': routing_key}

    def declare_task(self, name, taskgroup=None):
        """Declares a task without providing its implementation function. This
        may be used by client applications that know a task exists and how to
        use it, but do not have (or need) access to the task function itself.

        :param name: The full name of the task.
        :type name: str
        :param taskgroup: The taskgroup to use for routing the task messages.
                          Must have been previously declared with
                          :meth:`.declare_taskgroup`. Default taskgroup uses
                          exchange ``''`` and the task name as the routing key.
        :type taskgroup: str

        """
        tg_info = self._taskgroups[taskgroup]
        self.tasks._declare(name, **tg_info)

    def register_task(self, func, name=None, taskgroup=None):
        """Registers a single function as an available task.

        :param func: The function to call when executing the task.
        :type func: collections.Callable
        :param name: A string to uniquely identify the task, ``func.__name__``
                     is used by default.
        :type name: str
        :param taskgroup: An optional, previously declared taskgroup name to
                          make the task a member of. This overrides the use of
                          the :func:`~provoke.app.taskgroup` decorator.
        :type taskgroup: str

        """
        if name is None:
            name = func.__name__
        tg_info = {'exchange': None, 'routing_key': None}
        if taskgroup is not None:
            tg_info = self._taskgroups[taskgroup]
        elif hasattr(func, '_taskgroup'):
            tg_info = self._taskgroups[func._taskgroup]
        self.tasks._set(func, name, **tg_info)

    def register_module(self, mod, prefix='', taskgroup=None, what=None):
        """Convenience method for bulk registering tasks in a module.

        The module may use a special attribute named ``__declare_tasks__`` to
        specify a list of ``(name, taskgroup)`` tuples, each tuple is passed
        into a call to :meth:`.declare_task`.

        :param mod: The module object to register tasks from.
        :type mod: module
        :param prefix: String prefixed to each task registered from the module.
        :type prefix: str
        :param taskgroup: An optional, previously declared taskgroup name to
                          make the task a member of. This overrides the use of
                          the :func:`~provoke.app.taskgroup` decorator.
        :type taskgroup: str
        :param what: List of module attribute names to use as tasks. By
                     default, ``mod.__all__`` is used.

        """
        if what is None:
            what = mod.__all__
        for func_identifier in what:
            func = getattr(mod, func_identifier)
            func_name = prefix + func.__name__
            self.register_task(func, func_name, taskgroup)
        for taskgroup, name in getattr(mod, '__declare_tasks__', []):
            self.declare_task(name, taskgroup)




import unittest

try:
    from mock import patch, MagicMock
except ImportError:
    from unittest.mock import patch, MagicMock
import amqp

from provoke.amqp import AmqpConnection, _PoolableAmqp
from provoke.connectionpool import ConnectionPool


@patch.object(ConnectionPool, 'enabled', new=False)
class TestPoolableAmqp(unittest.TestCase):

    @patch.object(amqp, 'Connection')
    def test_check(self, conn_class_mock):
        conn_class_mock.return_value = conn = MagicMock()
        ctx = _PoolableAmqp(one=1)
        self.assertTrue(ctx.check())
        conn_class_mock.assert_called_with(one=1)
        conn.send_heartbeat.assert_called_with()

    @patch.object(amqp, 'Connection')
    def test_check_fail(self, conn_class_mock):
        conn_class_mock.return_value = conn = MagicMock()
        conn.send_heartbeat.side_effect = ValueError
        ctx = _PoolableAmqp(one=1)
        self.assertFalse(ctx.check())
        conn_class_mock.assert_called_with(one=1)
        conn.send_heartbeat.assert_called_with()

    @patch.object(amqp, 'Connection')
    def test_close(self, conn_class_mock):
        conn_class_mock.return_value = conn = MagicMock()
        ctx = _PoolableAmqp(one=1)
        ctx.close()
        conn_class_mock.assert_called_with(one=1)
        conn.close.assert_called_with()

    @patch.object(amqp, 'Connection')
    def test_close_fd(self, conn_class_mock):
        conn_class_mock.return_value = conn = MagicMock()
        conn.close.side_effect = ValueError
        ctx = _PoolableAmqp(one=1)
        self.assertRaises(ValueError, ctx.close)
        conn_class_mock.assert_called_with(one=1)
        conn.sock.close.assert_called_with()


@patch.object(ConnectionPool, 'enabled', new=False)
class TestAmqpConnection(unittest.TestCase):

    @patch.object(amqp, 'Connection')
    def test_set_connection_params(self, conn_class_mock):
        old = AmqpConnection._connection_params
        AmqpConnection.set_connection_params(host='testhost',
                                             username='testuser',
                                             password='testpass')
        self.assertEqual({'host': 'testhost',
                          'username': 'testuser',
                          'password': 'testpass'},
                         AmqpConnection._connection_params)
        conn_class_mock.return_value = MagicMock()
        AmqpConnection._connection_params = old

    @patch.object(amqp, 'Connection')
    def test_successful(self, conn_class_mock):
        conn_class_mock.return_value = conn_mock = MagicMock()
        conn_mock.channel.return_value = chan_mock = MagicMock()
        with AmqpConnection() as channel:
            self.assertEqual(chan_mock, channel)
        conn_class_mock.assert_called_with()
        conn_mock.channel.assert_called_with()
        chan_mock.close.assert_called_with()
        conn_mock.close.assert_called_with()

    @patch.object(amqp, 'Connection')
    def test_exception_raised(self, conn_class_mock):
        conn_class_mock.return_value = conn_mock = MagicMock()
        conn_mock.channel.return_value = chan_mock = MagicMock()
        try:
            with AmqpConnection() as channel:
                self.assertEqual(chan_mock, channel)
                raise ValueError('failure')
        except ValueError as exc:
            self.assertEqual('failure', str(exc))
        else:
            self.fail('ValueError was suppressed or not raised')
        conn_class_mock.assert_called_with()
        conn_mock.channel.assert_called_with()
        chan_mock.close.assert_called_with()
        conn_mock.close.assert_called_with()

    @patch.object(amqp, 'Connection')
    def test_close_exception(self, conn_class_mock):
        conn_class_mock.return_value = conn_mock = MagicMock()
        conn_mock.channel.return_value = chan_mock = MagicMock()
        conn_mock.close.side_effect = ValueError('close failure')
        try:
            with AmqpConnection() as channel:
                self.assertEqual(chan_mock, channel)
        except ValueError as exc:
            self.assertEqual('close failure', str(exc))
        else:
            self.fail('ValueError was suppressed or not raised')
        conn_class_mock.assert_called_with()
        conn_mock.channel.assert_called_with()
        chan_mock.close.assert_called_with()
        conn_mock.close.assert_called_with()

    @patch.object(amqp, 'Connection')
    def test_disconnect_exception(self, conn_class_mock):
        conn_class_mock.return_value = conn_mock = MagicMock()
        conn_mock.channel.return_value = chan_mock = MagicMock()
        conn_mock.disconnect.side_effect = ValueError('disconnect failure')
        with AmqpConnection() as channel:
            self.assertEqual(chan_mock, channel)
        conn_class_mock.assert_called_with()
        conn_mock.channel.assert_called_with()
        chan_mock.close.assert_called_with()
        conn_mock.close.assert_called_with()



from __future__ import print_function

import os

from provoke.app import WorkerApplication
from provoke.amqp import AmqpConnection


def do_work(*received):
    result = list(reversed(received))
    print('===> PID {0}: doing work!'.format(os.getpid()))
    print('Received:', ' '.join(received))
    print('Result:', ' '.join(result))
    print('===> PID {0}: done!'.format(os.getpid()))
    return result


def register(master, config):
    with AmqpConnection() as channel:
        channel.queue_declare('do_work')

    app = WorkerApplication()
    app.register_task(do_work)
    master.add_worker(app, ['do_work'], num_processes=4)




import unittest

try:
    from mock import patch, MagicMock
except ImportError:
    from unittest.mock import patch, MagicMock

try:
    import MySQLdb as mysql_mod
    connect_func = 'connect'
except ImportError:
    import pymysql as mysql_mod
    connect_func = 'Connect'

from provoke.mysql import MySQLConnection, _MySQLContext
from provoke.connectionpool import ConnectionPool


@patch.object(ConnectionPool, 'enabled', new=False)
class TestMySQLContext(unittest.TestCase):

    @patch.object(mysql_mod, connect_func)
    def test_check(self, conn_func_mock):
        conn_func_mock.return_value = conn = MagicMock()
        ctx = _MySQLContext(one=1)
        self.assertTrue(ctx.check())
        conn_func_mock.assert_called_with(one=1)
        conn.ping.assert_called_with()

    @patch.object(mysql_mod, connect_func)
    def test_check_fail(self, conn_func_mock):
        conn_func_mock.return_value = conn = MagicMock()
        conn.ping.side_effect = mysql_mod.OperationalError
        ctx = _MySQLContext(one=1)
        self.assertFalse(ctx.check())
        conn_func_mock.assert_called_with(one=1)
        conn.ping.assert_called_with()

    @patch.object(mysql_mod, connect_func)
    def test_close(self, conn_func_mock):
        conn_func_mock.return_value = conn = MagicMock()
        ctx = _MySQLContext(one=1)
        ctx.close()
        conn_func_mock.assert_called_with(one=1)
        conn.close.assert_called_with()


@patch.object(ConnectionPool, 'enabled', new=False)
class TestMySQLConnection(unittest.TestCase):

    def test_set_connection_params(self):
        old = MySQLConnection._connection_params.copy()
        expected = {'host': 'testhost',
                    'user': 'testuser',
                    'password': 'testpass'}
        MySQLConnection.set_connection_params('test', **expected)
        received = MySQLConnection._connection_params.get('test')
        self.assertEqual(expected, received)
        MySQLConnection._connection_params = old

    @patch.object(mysql_mod, connect_func)
    def test_successful(self, conn_func_mock):
        conn_func_mock.return_value = conn_mock = MagicMock()
        with MySQLConnection('test') as testdb:
            self.assertEqual(conn_mock, testdb.conn)
        conn_func_mock.assert_called_with()
        conn_mock.close.assert_called_with()

    @patch.object(mysql_mod, connect_func)
    def test_exception_raised(self, conn_func_mock):
        conn_func_mock.return_value = conn_mock = MagicMock()
        try:
            with MySQLConnection('test') as testdb:
                self.assertEqual(conn_mock, testdb.conn)
                raise ValueError('failure')
        except ValueError as exc:
            self.assertEqual('failure', str(exc))
        else:
            self.fail('ValueError was suppressed or not raised')
        conn_func_mock.assert_called_with()
        conn_mock.close.assert_called_with()

    @patch.object(mysql_mod, connect_func)
    def test_close_exception(self, conn_func_mock):
        conn_func_mock.return_value = conn_mock = MagicMock()
        conn_mock.close.side_effect = ValueError('close failure')
        try:
            with MySQLConnection('test') as testdb:
                self.assertEqual(conn_mock, testdb.conn)
        except ValueError as exc:
            self.assertEqual('close failure', str(exc))
        else:
            self.fail('ValueError was suppressed or not raised')
        conn_func_mock.assert_called_with()
        conn_mock.close.assert_called_with()




import unittest

try:
    from mock import patch, MagicMock
except ImportError:
    from unittest.mock import patch, MagicMock
from six.moves import http_client

from provoke.http import HttpConnection


class TestHttpConnection(unittest.TestCase):

    def setUp(self):
        HttpConnection._connection_params = {}

    def test_reset_connection_params(self):
        HttpConnection.set_connection_params('test', host='testhost')
        HttpConnection.reset_connection_params()
        self.assertEqual({}, HttpConnection._connection_params)

    def test_set_connection_params(self):
        HttpConnection.set_connection_params('test',
                                             host='testhost',
                                             user='testuser',
                                             password='testpass',
                                             ssl=True)
        self.assertEqual({'test': {'host': 'testhost',
                                   'user': 'testuser',
                                   'password': 'testpass',
                                   'ssl': True}},
                         HttpConnection._connection_params)

    @patch.object(http_client, 'HTTPConnection')
    def test_successful(self, conn_class_mock):
        conn_class_mock.return_value = conn_mock = MagicMock()
        with HttpConnection('test') as (conn, headers):
            self.assertEqual(conn_mock, conn)
        conn_class_mock.assert_called_with('localhost')
        conn_mock.close.assert_called_with()

    @patch.object(http_client, 'HTTPSConnection')
    def test_successful_ssl(self, conn_class_mock):
        HttpConnection.set_connection_params('test', ssl=True)
        conn_class_mock.return_value = conn_mock = MagicMock()
        with HttpConnection('test') as (conn, headers):
            self.assertEqual(conn_mock, conn)
        conn_class_mock.assert_called_with('localhost')
        conn_mock.close.assert_called_with()

    @patch.object(http_client, 'HTTPConnection')
    def test_successful_auth(self, conn_class_mock):
        HttpConnection.set_connection_params('test',
                                             user='testuser',
                                             password='testpass')
        conn_class_mock.return_value = conn_mock = MagicMock()
        with HttpConnection('test') as (conn, headers):
            self.assertEqual(conn_mock, conn)
            expected = {'Authorization': b'Basic dGVzdHVzZXI6dGVzdHBhc3M='}
            self.assertEqual(expected, headers)
        conn_class_mock.assert_called_with('localhost')
        conn_mock.close.assert_called_with()

    @patch.object(http_client, 'HTTPConnection')
    def test_exception_raised(self, conn_class_mock):
        conn_class_mock.return_value = conn_mock = MagicMock()
        try:
            with HttpConnection('test') as (conn, headers):
                self.assertEqual(conn_mock, conn)
                raise ValueError('failure')
        except ValueError as exc:
            self.assertEqual('failure', str(exc))
        else:
            self.fail('ValueError was suppressed or not raised')
        conn_class_mock.assert_called_with('localhost')
        conn_mock.close.assert_called_with()

    @patch.object(http_client, 'HTTPConnection')
    def test_close_exception(self, conn_class_mock):
        conn_class_mock.return_value = conn_mock = MagicMock()
        conn_mock.close.side_effect = ValueError('close failure')
        try:
            with HttpConnection('test') as (conn, headers):
                self.assertEqual(conn_mock, conn)
        except ValueError as exc:
            self.assertEqual('close failure', str(exc))
        else:
            self.fail('ValueError was suppressed or not raised')
        conn_class_mock.assert_called_with('localhost')
        conn_mock.close.assert_called_with()

    @patch.object(http_client, 'HTTPConnection')
    def test_disconnect_exception(self, conn_class_mock):
        conn_class_mock.return_value = conn_mock = MagicMock()
        conn_mock.disconnect.side_effect = ValueError('disconnect failure')
        with HttpConnection('test') as (conn, headers):
            self.assertEqual(conn_mock, conn)
        conn_class_mock.assert_called_with('localhost')
        conn_mock.close.assert_called_with()




import unittest
import sys
import logging
import logging.config
import resource
from six.moves.configparser import NoOptionError

try:
    from mock import patch, MagicMock, ANY
except ImportError:
    from unittest.mock import patch, MagicMock, ANY

from provoke.config import Configuration
from provoke.app import WorkerApplication
from provoke.amqp import AmqpConnection
from provoke.mysql import MySQLConnection
from provoke.http import HttpConnection


class TestConfiguration(unittest.TestCase):

    def test_from_config_str(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.get.return_value = 'three'
        ret = {}
        cfg._from_config(ret, 'sec', 'onetest', dict_key='one')
        cfgparser.get.assert_called_with('sec', 'onetest')
        self.assertEqual('three', ret['one'])

    def test_from_config_missing(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.get.side_effect = NoOptionError('sec', 'one')
        ret = {}
        cfg._from_config(ret, 'sec', 'one')
        cfgparser.get.assert_called_with('sec', 'one')
        self.assertFalse('one' in ret)

    def test_from_config_int(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.getint.return_value = 13
        ret = {}
        cfg._from_config(ret, 'sec', 'one', opt_type='int')
        cfgparser.getint.assert_called_with('sec', 'one')
        self.assertEqual(13, ret['one'])

    def test_from_config_float(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.getfloat.return_value = 3.14
        ret = {}
        cfg._from_config(ret, 'sec', 'one', opt_type='float')
        cfgparser.getfloat.assert_called_with('sec', 'one')
        self.assertEqual(3.14, ret['one'])

    def test_from_config_bool(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.getboolean.return_value = True
        ret = {}
        cfg._from_config(ret, 'sec', 'one', opt_type='bool')
        cfgparser.getboolean.assert_called_with('sec', 'one')
        self.assertEqual(True, ret['one'])

    def test_from_config_list(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.get.return_value = 'one, two,\tthree'
        ret = {}
        cfg._from_config(ret, 'sec', 'one', opt_type='list')
        cfgparser.get.assert_called_with('sec', 'one')
        self.assertEqual(['one', 'two', 'three'], ret['one'])

    @patch.object(logging, 'basicConfig')
    def test_configure_logging(self, basicconfig_mock):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.get.side_effect = NoOptionError(None, None)
        cfg.configure_logging()
        cfgparser.get.assert_called_once_with('daemon', 'logging_config')
        basicconfig_mock.assert_called_once_with(
            stream=sys.stdout, format=ANY, level=ANY)

    @patch.object(logging.config, 'fileConfig')
    def test_configure_logging_file(self, fileconfig_mock):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.get.return_value = '/path/to/logging.conf'
        cfg.configure_logging()
        cfgparser.get.assert_called_once_with('daemon', 'logging_config')
        fileconfig_mock.assert_called_once_with('/path/to/logging.conf')

    @patch.object(MySQLConnection, 'reset_connection_params')
    @patch.object(MySQLConnection, 'set_connection_params')
    def test_configure_mysql(self, set_mock, reset_mock):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.sections.return_value = ['one', 'mysql:test']
        cfgparser.get.side_effect = [
            'testuser',
            'testpass',
            'testhost',
            'testdb',
            'testcharset',
            NoOptionError('mysql:test', 'unix_socket')]
        cfgparser.getint.side_effect = [3306, 10]
        cfg.configure_mysql()
        reset_mock.assert_called_with()
        set_mock.assert_called_with(
            'test',
            user='testuser',
            password='testpass',
            host='testhost',
            port=3306,
            database='testdb',
            charset='testcharset',
            connect_timeout=10)

    @patch.object(HttpConnection, 'reset_connection_params')
    @patch.object(HttpConnection, 'set_connection_params')
    def test_configure_http(self, set_mock, reset_mock):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.sections.return_value = ['one', 'http:test']
        cfgparser.get.side_effect = ['testhost',
                                     'testuser',
                                     'testpass',
                                     NoOptionError(None, None),
                                     NoOptionError(None, None)]
        cfgparser.getint.side_effect = [3306,
                                        NoOptionError(None, None)]
        cfgparser.getboolean.side_effect = [True]
        cfg.configure_http()
        reset_mock.assert_called_with()
        set_mock.assert_called_with('test',
                                    host='testhost',
                                    port=3306,
                                    user='testuser',
                                    password='testpass',
                                    ssl=True)

    @patch.object(AmqpConnection, 'set_connection_params')
    def test_configure_amqp(self, set_mock):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.sections.return_value = ['one', 'amqp']
        cfgparser.get.side_effect = [
            'testhost',
            'testuser',
            'testpass',
            NoOptionError('amqp', 'virtual_host')]
        cfgparser.getint.return_value = 5672
        cfgparser.getfloat.side_effect = [30.0, 10.0]
        cfg.configure_amqp()
        set_mock.assert_called_with(
            userid='testuser',
            password='testpass',
            host='testhost',
            port=5672,
            heartbeat=30.0,
            connect_timeout=10.0)

    @patch.object(WorkerApplication, 'reset_taskgroups')
    @patch.object(WorkerApplication, 'declare_taskgroup')
    def test_configure_taskgroups(self, declare_mock, reset_mock):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.sections.return_value = ['one', 'taskgroup:testgroup']
        cfgparser.get.side_effect = [NoOptionError(None, 'queue'),
                                     'testkey',
                                     'testexchange']
        cfg.configure_taskgroups()
        reset_mock.assert_called_with()
        declare_mock.assert_called_with('testgroup', exchange='testexchange',
                                        routing_key='testkey')

    def test_get_rlimits(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.getint.side_effect = [65535,
                                        NoOptionError('daemon', 'max-fd')]
        self.assertEqual([(resource.RLIMIT_NOFILE, (65535, 65535))],
                         list(cfg.get_rlimits()))
        cfgparser.getint.assert_called_with('daemon', 'max-fd')
        self.assertEqual([], list(cfg.get_rlimits()))
        cfgparser.getint.assert_called_with('daemon', 'max-fd')

    def test_get_pidfile(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.get.side_effect = ['/var/run/test',
                                     NoOptionError('daemon', 'pidfile')]
        self.assertEqual('/var/run/test', cfg.get_pidfile())
        cfgparser.get.assert_called_with('daemon', 'pidfile')
        self.assertEqual(None, cfg.get_pidfile())
        cfgparser.get.assert_called_with('daemon', 'pidfile')

    def test_get_stdio_redirects(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.get.side_effect = ['test_stdout',
                                     'test_stderr',
                                     'test_stdin']
        self.assertEqual(('test_stdout', 'test_stderr', 'test_stdin'),
                         cfg.get_stdio_redirects())
        self.assertEqual(3, cfgparser.get.call_count)
        cfgparser.get.assert_any_call('daemon', 'stdout')
        cfgparser.get.assert_any_call('daemon', 'stderr')
        cfgparser.get.assert_any_call('daemon', 'stdin')

    def test_get_worker_privileges(self):
        cfgparser = MagicMock()
        cfg = Configuration(cfgparser)
        cfgparser.get.side_effect = ['test_user',
                                     'test_group',
                                     '1234']
        self.assertEqual(('test_user', 'test_group', 1234),
                         cfg.get_worker_privileges())
        self.assertEqual(3, cfgparser.get.call_count)
        cfgparser.get.assert_any_call('daemon', 'user')
        cfgparser.get.assert_any_call('daemon', 'group')
        cfgparser.get.assert_any_call('daemon', 'umask')



# The MIT License (MIT)
#
# Copyright (c) 2014 Ian Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""This module defines a set of routines for gathering needed information from
a config file.

"""

from __future__ import absolute_import

import sys
import re
import logging
import logging.config
import resource
from ast import literal_eval
from six.moves.configparser import NoOptionError, NoSectionError

__all__ = ['Configuration']


_COMMA_DELIMITER = re.compile(r',\s*')


class Configuration(object):
    """This class streamlines the process of reading data from config files.

    :param config: The config parser object.
    :type config: :py:class:`~ConfigParser.SafeConfigParser`

    """

    def __init__(self, config):
        super(Configuration, self).__init__()
        self._config = config

    def _from_config(self, in_dict, section, name, dict_key=None,
                     opt_type='str'):
        config = self._config
        if dict_key is None:
            dict_key = name
        try:
            if opt_type == 'int':
                in_dict[dict_key] = config.getint(section, name)
            elif opt_type == 'float':
                in_dict[dict_key] = config.getfloat(section, name)
            elif opt_type == 'bool':
                in_dict[dict_key] = config.getboolean(section, name)
            elif opt_type == 'list':
                values = config.get(section, name)
                in_dict[dict_key] = re.split(_COMMA_DELIMITER, values)
            else:
                in_dict[dict_key] = config.get(section, name)
        except (NoOptionError, NoSectionError):
            pass

    def configure_logging(self, section='daemon'):
        """Looks for a logging config file and, if it finds one, configures
        logging using :py:func:`logging.config.fileConfig`. Otherwise, uses
        :py:func:`logging.basicConfig` to standard output.

        """
        try:
            log_config = self._config.get(section, 'logging_config')
        except (NoSectionError, NoOptionError):
            format = '%(asctime)-26s %(levelname)-8s %(name)-14s %(message)s'
            logging.basicConfig(stream=sys.stdout, format=format,
                                level=logging.WARN)
        else:
            logging.config.fileConfig(log_config)

    def configure_mysql(self):
        """Searches for config sections prefixed with ``mysql:`` and loads them
        into the global dictionary of MySQL databases.

        The section name after the prefix will be the name of the database for
        later use.

        """
        try:
            from .mysql import MySQLConnection
        except ImportError:
            return
        MySQLConnection.reset_connection_params()
        section_prefix = 'mysql:'
        for section in self._config.sections():
            if section.startswith(section_prefix):
                db_name = section[len(section_prefix):]
                params = {}
                self._from_config(params, section, 'user')
                self._from_config(params, section, 'password')
                self._from_config(params, section, 'host')
                self._from_config(params, section, 'port', opt_type='int')
                self._from_config(params, section, 'database')
                self._from_config(params, section, 'charset')
                self._from_config(params, section, 'unix_socket')
                self._from_config(params, section, 'connect_timeout',
                                  opt_type='int')
                MySQLConnection.set_connection_params(db_name, **params)

    def configure_http(self):
        """Searches for config sections prefixed with ``http:`` and loads them
        into the global dictionary of HTTP endpoints.

        The section name after the prefix will be the name of the endpoint for
        later use.

        """
        try:
            from .http import HttpConnection
        except ImportError:
            return
        HttpConnection.reset_connection_params()
        section_prefix = 'http:'
        for section in self._config.sections():
            if section.startswith(section_prefix):
                name = section[len(section_prefix):]
                params = {}
                self._from_config(params, section, 'host')
                self._from_config(params, section, 'port', opt_type='int')
                self._from_config(params, section, 'user')
                self._from_config(params, section, 'password')
                self._from_config(params, section, 'timeout', opt_type='int')
                self._from_config(params, section, 'ssl', opt_type='bool')
                self._from_config(params, section, 'key_file')
                self._from_config(params, section, 'cert_file')
                HttpConnection.set_connection_params(name, **params)

    def configure_amqp(self):
        """Loads AMQP connection information from the config section ``amqp``.
        This connection information will be available globally.

        If this section does not exist, or if this method is not called, the
        default connection parameters will look for a server running on
        localhost.

        """
        try:
            from .amqp import AmqpConnection
        except ImportError:
            return
        AmqpConnection.set_connection_params()
        section = 'amqp'
        if self._config.has_section(section):
            params = {}
            self._from_config(params, section, 'host')
            self._from_config(params, section, 'port', opt_type='int')
            self._from_config(params, section, 'user', dict_key='userid')
            self._from_config(params, section, 'password')
            self._from_config(params, section, 'virtual_host')
            self._from_config(params, section, 'heartbeat', opt_type='float')
            self._from_config(params, section, 'connect_timeout',
                              opt_type='float')
            AmqpConnection.set_connection_params(**params)

    def configure_taskgroups(self):
        """Searches for sections prefixed with ``taskgroup:`` and loads them
        into the global dictionary of application taskgroups.

        """
        try:
            from .app import WorkerApplication
        except ImportError:
            return
        WorkerApplication.reset_taskgroups()
        section_prefix = 'taskgroup:'
        for section in self._config.sections():
            if section.startswith(section_prefix):
                tg_name = section[len(section_prefix):]
                params = {'exchange': '',
                          'routing_key': None}
                self._from_config(params, section, 'queue')
                self._from_config(params, section, 'routing_key')
                self._from_config(params, section, 'exchange')
                if 'queue' in params:
                    params['exchange'] = ''
                    params['routing_key'] = params.pop('queue')
                WorkerApplication.declare_taskgroup(tg_name, **params)

    def get_rlimits(self, section='daemon'):
        try:
            max_fd = self._config.getint(section, 'max-fd')
            yield resource.RLIMIT_NOFILE, (max_fd, max_fd)
        except (NoOptionError, NoSectionError):
            pass

    def get_pidfile(self, section='daemon'):
        try:
            return self._config.get(section, 'pidfile')
        except (NoOptionError, NoSectionError):
            pass

    def get_stdio_redirects(self, section='daemon'):
        info = {}
        self._from_config(info, section, 'stdout')
        self._from_config(info, section, 'stderr')
        self._from_config(info, section, 'stdin')
        return info.get('stdout'), info.get('stderr'), info.get('stdin')

    def get_worker_privileges(self, section='daemon'):
        info = {}
        self._from_config(info, section, 'user')
        self._from_config(info, section, 'group')
        self._from_config(info, section, 'umask')
        if 'umask' in info:
            info['umask'] = literal_eval(info['umask'])
        return info.get('user'), info.get('group'), info.get('umask')



# The MIT License (MIT)
#
# Copyright (c) 2014 Ian Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Worker master module. Provides a system that maintains a set of worker
processes that consume from task queues. If a worker process dies for any
reason, a new one is created.

"""

from __future__ import absolute_import

import os
import time
import json
import logging
import errno
import signal
import traceback
from socket import timeout as socket_timeout
from functools import partial

from six.moves import cPickle
import amqp
from amqp.exceptions import AccessRefused

from ..amqp import AmqpConnection

__all__ = ['WorkerMaster', 'DiscardTask', 'RequeueTask',
           'get_worker_data', 'get_worker_app']


_current_worker_data = None
_current_worker_app = None

logger = logging.getLogger('provoke.worker')


def get_worker_data(key, default=None):
    """This function, when called from within tasks running in worker
    processes, will get values from the ``worker_data`` dictionary passed in to
    the :class:`WorkerMaster` constructor.

    :param key: The key to retrieve from the ``worker_data`` dictionary.
    :param default: This value is returned when ``key`` cannot be found in
                    ``worker_data``.
    :returns: The value stored for ``key`` in the ``worker_data`` dictionary.
    :raises: RuntimeError

    """
    if _current_worker_data:
        return _current_worker_data.get(key, default)
    else:
        raise RuntimeError('Must be called from worker process')


def get_worker_app():
    """This function, when called from within tasks running in worker
    processes, will get the :class:`~provoke.app.WorkerApplication` object that
    is executing the task.

    :rtype: :class:`provoke.app.WorkerApplication`
    :raises: RuntimeError

    """
    if _current_worker_app:
        return _current_worker_app
    else:
        raise RuntimeError('Must be called from worker process')


class DiscardTask(Exception):
    """May be raised from the callback that runs before each task to indicate
    that the task should be discarded without execution. The task is
    acknowledged to the AMQP server as if it were successfully processed.

    The ``return_callback`` that runs after task execution will not be called
    if the task is discarded by raising this exception.

    """
    pass


class RequeueTask(Exception):
    """May be called from inside task functions to indicate that the AMQP
    message should be rejected but requeued immediately. This should
    effectively initiate a retry of the task.

    To prevent an infinite loop of requeuing, tasks should use a guard against
    requeuing messages that have already been requeued::

        if not get_worker_data('redelivered'):
            raise RequeueTask()

    """
    pass


class _WorkerProcess(object):

    def __init__(self, app, queues, limit, task_callback,
                 return_callback, exclusive):
        super(_WorkerProcess, self).__init__()
        self.app = app
        self.queues = queues
        self.limit = limit
        self.task_callback = task_callback
        self.return_callback = return_callback
        self.exclusive = exclusive
        self.pid = None

    def _send_result(self, channel, reply_to, body):
        result_raw = json.dumps(body)
        msg = amqp.Message(result_raw, content_type='application/json')
        channel.basic_publish(msg, exchange='', routing_key=reply_to)

    def _handle_message(self, channel, msg):
        body = json.loads(msg.body)
        task_name = body['task']
        task_args = body.get('args', [])
        task_kwargs = body.get('kwargs')
        task_id = getattr(msg, 'correlation_id', None)
        reply_to = getattr(msg, 'reply_to', None)
        call = getattr(self.app.tasks, task_name)
        skip = False
        _current_worker_data['correlation_id'] = task_id
        _current_worker_data['redelivered'] = \
            msg.delivery_info.get('redelivered')
        if self.task_callback:
            try:
                self.task_callback(task_name, task_args, task_kwargs)
            except DiscardTask:
                skip = True
        if not skip:
            try:
                ret = call.apply(task_args, task_kwargs, task_id)
            except Exception as exc:
                if reply_to:
                    body['exception'] = {'value': cPickle.dumps(exc),
                                         'traceback': traceback.format_exc()}
                    self._send_result(channel, reply_to, body)
                if self.return_callback:
                    self.return_callback(task_name, None)
                raise
            else:
                if reply_to:
                    body['return'] = ret
                    self._send_result(channel, reply_to, body)
                if self.return_callback:
                    self.return_callback(task_name, ret)

    def _on_message(self, channel, msg):
        try:
            self._handle_message(channel, msg)
        except (SystemExit, KeyboardInterrupt):
            channel.basic_reject(msg.delivery_tag, requeue=True)
            raise
        except RequeueTask:
            channel.basic_reject(msg.delivery_tag, requeue=True)
        except Exception:
            channel.basic_reject(msg.delivery_tag, requeue=False)
            raise
        else:
            channel.basic_ack(msg.delivery_tag)
        finally:
            self.counter += 1
            if self.limit and self.counter >= self.limit:
                self._cancel_consumers(channel)

    def _cancel_consumers(self, channel):
        for consumer_tag in channel.callbacks.keys():
            channel.basic_cancel(consumer_tag)

    def _consume(self):
        self.counter = 0

        with AmqpConnection() as channel:
            callback = partial(self._on_message, channel)
            channel.basic_qos(0, 1, False)
            for queue_name in self.queues:
                channel.basic_consume(queue=queue_name,
                                      consumer_tag=queue_name,
                                      callback=callback,
                                      exclusive=self.exclusive)
            logger.info('Accepting jobs: queues=%s', repr(self.queues))
            while channel.callbacks:
                try:
                    channel.connection.drain_events(timeout=10.0)
                except socket_timeout:
                    channel.connection.send_heartbeat()

    def _try_consuming(self):
        while True:
            try:
                self._consume()
                return
            except AccessRefused:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception('Queue access refused')
                time.sleep(5.0)

    def _run(self):
        try:
            self._try_consuming()
        except (SystemExit, KeyboardInterrupt):
            pass
        except Exception:
            logger.exception('Unhandled exception in worker process')
            traceback.print_exc()
            raise


class WorkerMaster(object):
    """Manages child processes that execute application workers. These workers
    may be listening on one or many queues.

    :param start_callback: This function is called in the master process every
                           time a new worker process is started. This callback
                           is given three parameters, the
                           :class:`~provoke.app.WorkerApplication`, a list of
                           queues consumed by the new process, and the PID of
                           the new process.
    :type start_callback: collections.Callable
    :param exit_callback: This function is called in the master process every
                          time a worker process exits for any reason. It is
                          passed in the same arguments as ``start_callback``
                          plus the exit status integer.
    :type exit_callback: collections.Callable
    :param process_callback: This function is called in new child processes
                             before any tasks are executed. It is passed in a
                             single list of queues that will be consumed in the
                             new process.
    :type process_callback: collections.Callable
    :param worker_data: Arbitrary data may be made available to workers with
                        this dictionary. Tasks running in worker processes may
                        use
                        :func:`~provoke.worker.master.get_worker_data` to
                        access copies of this dictionary.
    :type worker_data: dict

    """

    def __init__(self, start_callback=None, exit_callback=None,
                 process_callback=None, worker_data=None):
        super(WorkerMaster, self).__init__()
        self._start_callback = start_callback
        self._exit_callback = exit_callback
        self._process_callback = process_callback
        self._internal_process_callback = None
        self._worker_data = worker_data or {}
        self.workers = []

    def start_callback(self, worker):
        logger.info('Process started: pid=%s, queues=%s',
                    worker.pid, repr(worker.queues))
        if self._start_callback:
            try:
                self._start_callback(worker.queues, worker.pid)
            except Exception:
                pass

    def exit_callback(self, worker, status):
        logger.info('Process exited: pid=%s, status=%s, queues=%s',
                    worker.pid, status, repr(worker.queues))
        if self._exit_callback:
            try:
                self._exit_callback(worker.queues, worker.pid, status)
            except Exception:
                pass

    def process_callback(self, worker):
        if self._internal_process_callback:
            self._internal_process_callback()
        if self._process_callback:
            try:
                self._process_callback(worker.queues)
            except Exception:
                pass

    def add_worker(self, app, queues, num_processes=1, task_limit=10,
                   task_callback=None, return_callback=None, exclusive=False):
        """Adds a new worker process to be managed by the :meth:`.run` method.

        :param app: The application backend that knows how to enqueue and
                    execute tasks.
        :type app: :class:`~provoke.app.WorkerApplication`
        :param queues: List of queue names to consume task execution messages
                       from.
        :type queues: list
        :param num_processes: The number of processes to maintain for this
                              worker.
        :type num_processes: int
        :param task_limit: The number of tasks a worker process may execute
                           before it exits and is replaced by a new worker
                           process.
        :type task_limit: int
        :param task_callback: This function is called inside the child process
                              every time a task is ready for execution. This
                              function is given four arguments, the
                              :class:`~provoke.app.WorkerApplication`, the task
                              name, and the positional and keyword arguments of
                              the task.
        :type task_callback: collections.Callable
        :param return_callback: Like ``task_callback`` but called when the
                                task is completed. This function is given three
                                arguments, the
                                :class:`~provoke.app.WorkerApplication`, the
                                task name, and the return value. If an
                                exception was raised during execution, it will
                                be available in :func:`sys.exc_info`.
        :type return_callback: collections.Callable
        :param exclusive: If True, the worker process will request
                          exclusive access to consume the queues. Additional
                          consume requests on the queue by other workers will
                          raise errors. Only makes sense to use one process!
        :type exclusive: bool

        """
        for i in range(num_processes):
            worker = _WorkerProcess(app, queues, task_limit, task_callback,
                                    return_callback, exclusive)
            self.workers += [worker]

    def _check_workers(self):
        try:
            pid, status = os.waitpid(0, 0)
        except OSError as exc:
            if exc.errno == errno.ECHILD:
                for worker in self.workers:
                    self.exit_callback(worker, None)
                    worker.pid = None
                return False
            raise
        for worker in self.workers:
            exit_status = os.WEXITSTATUS(status)
            if pid == worker.pid:
                self.exit_callback(worker, exit_status)
                worker.pid = None
                return True
        raise Exception('Received exit status for unknown process: '+pid)

    def _start_worker(self, worker):
        pid = os.fork()
        if pid == 0:
            global _current_worker_data, _current_worker_app
            _current_worker_data = self._worker_data.copy()
            _current_worker_app = worker.app
            self.process_callback(worker)
            try:
                worker._run()
            except Exception:
                os._exit(1)
            finally:
                os._exit(0)
        else:
            return pid

    def _restart_workers(self):
        for worker in self.workers:
            if worker.pid is None:
                worker.pid = self._start_worker(worker)
                self.start_callback(worker)

    def _stop_workers(self):
        for worker in self.workers:
            if worker.pid is not None:
                try:
                    os.kill(worker.pid, signal.SIGTERM)
                except OSError as exc:
                    if exc.errno != errno.ESRCH:
                        raise

    def wait(self):
        """This method may be used after :meth:`.run` to wait for killed worker
        processes to finish and return an exit status. This is necessary for
        ensuring ``exit_callback`` is called appropriately::

            try:
                master.run()
            finally:
                master.wait()

        """
        for worker in self.workers:
            if worker.pid is not None:
                try:
                    pid, status = os.waitpid(worker.pid, 0)
                except OSError as exc:
                    if exc.errno in (errno.ECHILD, errno.ESRCH):
                        status = None
                    else:
                        raise
                self.exit_callback(worker, status)
                worker.pid = None

    def run(self):
        """Starts all worker processes, and as they exit they are restarted.
        When this function exits for any reason (e.g. ``KeyboardInterrupt``),
        ``SIGTERM`` is sent to each worker process.

        """
        try:
            while True:
                self._restart_workers()
                self._check_workers()
        finally:
            self._stop_workers()



# The MIT License (MIT)
#
# Copyright (c) 2014 Ian Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Common database module. Provides a central point where databases may be
configured, and a context manager to ensure connections are closed after use.

.. py:class:: _MySQLContext

   This is the object returned by the :class:`MySQLConnection` context manager.

   .. py:attribute:: conn

      Database connection object, conforming to the Python DB API 2.0. This
      connection is already open and will be closed when the context manager
      exits.

   .. py:attribute:: params

      Contains a dictionary of the parameters used to create the database
      connection.

   .. py:attribute:: module

      A reference to the module containing the DB API 2.0 interface that the
      connection was made with. This is useful for accessing exception classes.

"""

from __future__ import absolute_import

import logging

import pymysql

from ..connectionpool import ConnectionPool

__all__ = ['MySQLConnection']


logger = logging.getLogger('provoke.mysql')


class _MySQLContext(object):

    def __init__(self, **params):
        super(_MySQLContext, self).__init__()
        self.conn = pymysql.Connect(**params)
        self.module = pymysql
        self._set_session_vars()
        host = params.get('host', 'localhost')
        db = params.get('db')
        logger.debug('Connection established: host=%s, db=%s', host, db)

    def _set_session_vars(self):
        cur = self.conn.cursor()
        try:
            cur.execute("""SET SESSION `time_zone` = '+00:00'""")
            self.conn.commit()
        finally:
            cur.close()

    def check(self):
        try:
            self.conn.ping()
        except pymysql.OperationalError:
            return False
        return True

    def rollback(self):
        try:
            self.conn.rollback()
        except Exception:
            pass

    def close(self):
        self.conn.close()


class MySQLConnection(object):
    """Context manager that allows easy connection to databases by name using
    configured connection parameters. For example::

        MySQLConnection.set_connection_params('admin', host='...')

        with MySQLConnection('admin') as admin:
            try:
                cur = admin.conn.cursor()
                cur.execute("SELECT * FROM `tbl`")
                for row in cur:
                    print list(row)
            finally:
                cur.close()

    The connection is automatically closed when the context manager ends, but
    any cursors you created should be closed manually.

    :param db_names: The pre-configured database names to create connections
                     for. Upon context manager entry, one
                     :class:`_MySQLContext` object will be returned for each
                     entry in ``db_names``.
    :type db_names: str

    """

    _connection_params = {}
    _pools = None

    def __init__(self, *db_names):
        super(MySQLConnection, self).__init__()
        self.db_names = db_names
        self.db_conns = [None]*len(db_names)

    @classmethod
    def reset_connection_params(cls):
        """Removes all existing connection parameter information."""
        cls._connection_params = {}

    @classmethod
    def set_connection_params(cls, db_name, **params):
        """Configures future uses of the given name with a set of connection
        parameters.

        :param db_name: A unique, memorable name to reference the database by.
        :type db_name: str
        :param params: The database connection parameters, as you would pass in
                       to the ``.connect()`` function of the database.
        :type params: Keyword arguments

        """
        cls._connection_params[db_name] = params

    @classmethod
    def _get_pools(cls, db_names):
        if not cls._pools:
            cls._pools = {}
        for db_name in db_names:
            if db_name not in cls._pools:
                params = cls._connection_params.get(db_name, {})
                cls._pools[db_name] = ConnectionPool(_MySQLContext,
                                                     conn_kwargs=params)
        return cls._pools

    def __enter__(self):
        pools = self._get_pools(self.db_names)
        for i, db_name in enumerate(self.db_names):
            pool = pools[db_name]
            self.db_conns[i] = pool.get()
        if len(self.db_conns) > 1:
            return self.db_conns
        elif len(self.db_conns) == 1:
            return self.db_conns[0]

    def __exit__(self, exc_type, exc_val, exc_tb):
        pools = self._get_pools(self.db_names)
        try:
            for i, db_name in enumerate(self.db_names):
                self.db_conns[i].rollback()
                if exc_type:
                    self.db_conns[i].close()
                else:
                    pool = pools[db_name]
                    pool.release(self.db_conns[i])
        except Exception:
            if not exc_type:
                raise




import sys

from setuptools import setup, find_packages


if sys.version_info[0] == 2:
    mysql = 'MySQL-python'
else:
    mysql = 'PyMySQL'


setup(name='provoke',
      version='0.3.1',
      author='Ian Good',
      author_email='ian.good@rackspace.com',
      description='Lightweight, asynchronous function execution in Python '
                  'using AMQP.',
      packages=find_packages(),
      install_requires=['amqp', 'six'],
      extras_require={
          'mysql': [mysql],
      },
      entry_points={
          'provoke.workers': ['example = provoke.example.worker:register'],
          'console_scripts': ['provoke-worker = provoke.worker.main:main']})



# The MIT License (MIT)
#
# Copyright (c) 2014 Ian Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Common AMQP module. Provides a simple context manager interface for
connecting to an AMQP broker and returning a usable channel, closing both when
the context manager exits.

"""

from __future__ import absolute_import

import logging

import amqp

from .connectionpool import ConnectionPool

__all__ = ['AmqpConnection']


logger = logging.getLogger('provoke.amqp')


class _PoolableAmqp(object):

    def __init__(self, **kwargs):
        super(_PoolableAmqp, self).__init__()
        self.conn = amqp.Connection(**kwargs)
        host = kwargs.get('host', 'localhost')
        logger.debug('Connection established: host=%s', host)

    def check(self):
        try:
            self.conn.send_heartbeat()
        except Exception:
            return False
        return True

    def close(self):
        try:
            self.conn.close()
        except Exception:
            try:
                self.conn.sock.close()
            except Exception:
                pass
            raise


class AmqpConnection(object):
    """Defines the context manager for connecting to the AMQP broker and
    cleaning up afterwards. For example::

        with AmqpConnection() as channel:
            channel.basic_publish(exchange='',
                                  routing_key='testqueue',
                                  properties=my_properties,
                                  body='Test message')

    If custom connection parameters are needed (e.g. RabbitMQ is not running
    locally), they may be configured beforehand and all future context managers
    will use them.

    """

    _connection_params = {}
    _pool = None

    def __init__(self):
        super(AmqpConnection, self).__init__()

    @classmethod
    def set_connection_params(cls, **params):
        """Configures connection parameters that future usages of the context
        manager will use to connect to the AMQP broker.

        :param params: The AMQP connection parameters, as you would pass in to
                       the amqp's
                       :class:`~amqp.connection.Connection` class.
        :type params: Keyword arguments

        """
        cls._connection_params = params

    @classmethod
    def _get_pool(cls):
        if not cls._pool:
            cls._pool = ConnectionPool(_PoolableAmqp,
                                       conn_kwargs=cls._connection_params)
        return cls._pool

    def __enter__(self):
        pool = self._get_pool()
        self.conn = pool.get()
        self.channel = self.conn.conn.channel()
        return self.channel

    def __exit__(self, exc_type, exc_val, exc_tb):
        pool = self._get_pool()
        try:
            self.channel.close()
        except Exception:
            pass
        try:
            if exc_type:
                self.conn.close()
            else:
                pool.release(self.conn)
        except Exception:
            if not exc_type:
                raise




import unittest
import time
import json
from socket import timeout as socket_timeout
from multiprocessing import TimeoutError

try:
    from mock import patch, MagicMock, ANY
except ImportError:
    from unittest.mock import patch, MagicMock, ANY
import amqp

from provoke.amqp import AmqpConnection
from provoke.app import AsyncResult, _TaskCaller, _TaskSet, \
    taskgroup, WorkerApplication


class JsonMatcher(object):

    def __init__(self, test_case, expected):
        self.test_case = test_case
        self.expected = expected

    def __eq__(self, received):
        obj = json.loads(received)
        self.test_case.assertEqual(self.expected, obj)
        return True


class TestAsyncResult(unittest.TestCase):

    def test_attributes(self):
        res = AsyncResult('testid')
        self.assertEqual('testid', res.correlation_id)
        self.assertRaises(AttributeError, getattr, res, 'args')
        self.assertRaises(AttributeError, getattr, res, 'name')
        self.assertRaises(AttributeError, getattr, res, 'returned')
        self.assertRaises(AttributeError, getattr, res, 'exception')
        self.assertRaises(AttributeError, getattr, res, 'traceback')
        res._result = {'args': 123, 'kwargs': 456, 'task_name': 'test',
                       'exception': {'traceback': 'test traceback'}}
        self.assertEqual((123, 456), res.args)
        self.assertEqual('test', res.name)
        res._return = 789
        self.assertEqual(789, res.returned)
        del res._return
        res._exc = 987
        self.assertEqual(987, res.exception)
        self.assertEqual('test traceback', res.traceback)

    def test_on_message(self):
        res = AsyncResult(None)
        msg = MagicMock(body='{"task_name": "test", "return": 123}')
        res._on_message(msg)
        self.assertEqual(123, res._return)
        self.assertFalse(hasattr(res, '_exc'))

    def test_on_message_fail(self):
        res = AsyncResult(None)
        body = '{"task_name": "test", "exception": {"value": "I123\\n."}}'
        msg = MagicMock(body=body)
        res._on_message(msg)
        self.assertEqual(123, res._exc)
        self.assertFalse(hasattr(res, '_return'))

    def test_get_cached(self):
        res = AsyncResult(None)
        res._result = True
        res._return = 123
        self.assertEqual(123, res.get())
        del res._return
        res._exc = ValueError
        self.assertRaises(ValueError, res.get)

    @patch.object(AmqpConnection, '__enter__')
    @patch.object(AmqpConnection, '__exit__')
    def test_get(self, exit_mock, enter_mock):
        enter_mock.return_value = channel = MagicMock()
        exit_mock.return_value = None
        res = AsyncResult('test')

        def finish(timeout):
            res._result = True
            res._return = 123
        channel.connection.drain_events.side_effect = finish
        self.assertEqual(123, res.get())
        exit_mock.assert_called_with(None, None, None)
        channel.basic_consume.assert_called_with(queue='result_test',
                                                 no_ack=True, callback=ANY)
        channel.connection.drain_events.assert_called_with(timeout=10.0)
        channel.queue_delete.assert_called_with('result_test')
        self.assertFalse(channel.connection.send_heartbeat.called)

    @patch.object(AmqpConnection, '__enter__')
    @patch.object(AmqpConnection, '__exit__')
    @patch.object(AsyncResult, '_on_message')
    def test_get_nonblock(self, on_msg_mock, exit_mock, enter_mock):
        enter_mock.return_value = channel = MagicMock()
        exit_mock.return_value = None
        channel.basic_get.return_value = msg = MagicMock()
        res = AsyncResult('test')

        def finish(timeout):
            res._result = True
            res._return = 123
        on_msg_mock.side_effect = finish
        self.assertEqual(123, res.get(0.0))
        exit_mock.assert_called_with(None, None, None)
        on_msg_mock.assert_called_with(msg)
        channel.basic_get.assert_called_with(queue='result_test', no_ack=True)
        channel.queue_delete.assert_called_with('result_test')

    @patch.object(AmqpConnection, '__enter__')
    @patch.object(AmqpConnection, '__exit__')
    def test_get_keyerror(self, exit_mock, enter_mock):
        enter_mock.return_value = channel = MagicMock()
        exit_mock.return_value = None
        channel.basic_consume.side_effect = amqp.exceptions.NotFound
        res = AsyncResult('test')
        self.assertRaises(KeyError, res.get)
        channel.basic_consume.assert_called_with(queue='result_test',
                                                 no_ack=True, callback=ANY)
        self.assertFalse(channel.connection.drain_events.called)
        self.assertFalse(channel.queue_delete.called)
        self.assertFalse(channel.connection.send_heartbeat.called)

    @patch.object(AmqpConnection, '__enter__')
    @patch.object(AmqpConnection, '__exit__')
    @patch.object(time, 'time')
    def test_get_timeout(self, time_mock, exit_mock, enter_mock):
        enter_mock.return_value = channel = MagicMock()
        time_mock.side_effect = [0.0, 20.0]
        channel.connection.drain_events.side_effect = socket_timeout
        res = AsyncResult('test')
        self.assertRaises(TimeoutError, res.get, 10.0)
        channel.basic_consume.assert_called_with(queue='result_test',
                                                 no_ack=True, callback=ANY)
        channel.connection.drain_events.assert_called_with(timeout=0.0)
        channel.connection.send_heartbeat.assert_called_with()

    @patch.object(AsyncResult, 'get')
    def test_wait(self, get_mock):
        res = AsyncResult(None)
        res._return = True
        res.wait()
        del res._return
        get_mock.side_effect = ValueError
        res.wait()

    @patch.object(AsyncResult, 'wait')
    def test_ready(self, wait_mock):
        res = AsyncResult(None)
        self.assertFalse(res.ready())
        res._result = True
        self.assertTrue(res.ready())
        res._exc = True
        self.assertTrue(res.ready())
        wait_mock.assert_called_once_with(0.0)

    @patch.object(AsyncResult, 'ready')
    def test_successful(self, ready_mock):
        res = AsyncResult(None)
        ready_mock.return_value = False
        self.assertFalse(res.successful())
        ready_mock.return_value = True
        res._return = True
        self.assertTrue(res.successful())
        del res._return
        res._exc = True
        self.assertFalse(res.successful())


class TestTaskCall(unittest.TestCase):

    def test_delay(self):
        call = _TaskCaller('a', 'b', 'c', 'd', 'e')
        call.apply_async = MagicMock()
        call.delay('stuff', 13, test='what')
        call.apply_async.assert_called_with(('stuff', 13), {'test': 'what'})

    @patch.object(AmqpConnection, '__enter__')
    @patch.object(AmqpConnection, '__exit__')
    @patch.object(amqp, 'Message')
    def test_apply_async(self, msg_mock, amqp_exit_mock, amqp_enter_mock):
        channel = MagicMock()
        amqp_enter_mock.return_value = channel
        msg_mock.return_value = 52
        app = MagicMock(result_queue_ttl=0)
        func = MagicMock(_send_result=True)
        call = _TaskCaller(func, 'testname', app, 'testexchange', 'testqueue')
        ret = call.apply_async(('one', 'two'), {'three': 'four'},
                               send_result=True)
        self.assertTrue(isinstance(ret, AsyncResult))
        body_matcher = JsonMatcher(self, {'task': 'testname',
                                          'args': ['one', 'two'],
                                          'kwargs': {'three': 'four'}})
        msg_mock.assert_called_with(body_matcher,
                                    content_type='application/json',
                                    reply_to=ANY,
                                    correlation_id=ANY)
        amqp_enter_mock.assert_called_with()
        channel.basic_publish.assert_called_with(52, exchange='testexchange',
                                                 routing_key='testqueue')
        amqp_exit_mock.assert_called_with(None, None, None)

    def test_apply(self):
        func = MagicMock(return_value=13)
        call = _TaskCaller(func, 'a', 'b', 'c', 'd')
        ret = call.apply(('one', 'two'), {'three': 'four'})
        self.assertEqual(13, ret)
        func.assert_called_with('one', 'two', three='four')

    def test_call(self):
        func = MagicMock(return_value=13)
        call = _TaskCaller(func, 'a', 'b', 'c', 'd')
        ret = call('one', 'two', three='four')
        self.assertEqual(13, ret)
        func.assert_called_with('one', 'two', three='four')


class TestTaskSet(unittest.TestCase):

    def test_set(self):
        set = _TaskSet('testapp')
        set._set('testfunc', 'testname', 'testexchange', 'testrouting')
        self.assertEqual(1, len(set))
        self.assertTrue('testname' in set)
        self.assertEqual('testfunc', set._tasks['testname'].func)
        self.assertEqual('testname', set._tasks['testname'].name)
        self.assertEqual('testapp', set._tasks['testname'].app)
        self.assertEqual('testrouting', set._tasks['testname'].routing_key)
        self.assertEqual('testexchange', set._tasks['testname'].exchange)

    def test_declare(self):
        set = _TaskSet('testapp')
        set._set('testfunc1', 'testtask1', 'testexchange1', 'testrouting1')
        set._declare('testtask1', 'testexchange1', 'testrouting1')
        set._declare('testtask2', 'testexchange2', 'testrouting2')
        self.assertEqual(2, len(set))

    def test_getattr(self):
        set = _TaskSet('testparams')
        set._set('func1', 'one', 'testexchange', 'testrouting')
        set._set('func2', 'two', 'testexchange', 'testrouting')
        self.assertEqual('func1', set.one.func)
        self.assertEqual('func2', set.two.func)
        self.assertRaises(AttributeError, getattr, set, 'three')

    def test_repr(self):
        set = _TaskSet('testparams')
        set._set('func1', 'one', 'testexchange', 'testrouting')
        set._set('func2', 'two', 'testexchange', 'testrouting')
        self.assertTrue(repr(set).startswith('<registered task set '))
        self.assertTrue(repr(set).endswith('>'))


class TestWorkerApplication(unittest.TestCase):

    def test_declare_task(self):
        app = WorkerApplication()
        app.declare_taskgroup('testgroup', 'testexchange', 'testroutingkey')
        app.tasks = MagicMock()
        app.declare_task('taskname', 'testgroup')
        app.tasks._declare.assert_called_with('taskname',
                                              exchange='testexchange',
                                              routing_key='testroutingkey')

    def test_register_task(self):
        @taskgroup('testgroup')
        def func1():
            pass

        def func2():
            pass
        app = WorkerApplication()
        app.declare_taskgroup('testgroup', 'testexchange', 'testroutingkey')
        app.tasks = MagicMock()
        app.register_task(func1, 'funcone')
        app.tasks._set.assert_called_with(func1, 'funcone',
                                          exchange='testexchange',
                                          routing_key='testroutingkey')
        app.register_task(func2, taskgroup='testgroup')
        app.tasks._set.assert_called_with(func2, 'func2',
                                          exchange='testexchange',
                                          routing_key='testroutingkey')

    def test_register_module(self):
        app = WorkerApplication()
        app.declare_task = MagicMock()
        app.register_task = MagicMock()
        mod = MagicMock()
        mod.__all__ = ['func1', 'func2']
        mod.__declare_tasks__ = [('testgroup', 'testtask')]
        mod.func1 = MagicMock(__name__='func1')
        mod.func2 = MagicMock(__name__='func2')
        app.register_module(mod, 't_')
        app.register_task.assert_any_call(mod.func1, 't_func1', None)
        app.register_task.assert_any_call(mod.func2, 't_func2', None)
        app.declare_task.assert_called_with('testtask', 'testgroup')
